@@ -3,23 +3,27 @@ Definition of the base objects for the creation of a spec validator
 """
 import logging
 import os
+import pathlib
 from io import BytesIO
+from numbers import Integral
+from numbers import Real
 from pathlib import Path
 from pathlib import PurePath
 from typing import IO
 from typing import List
+from typing import Tuple
 from typing import Type
 from typing import Union
 
+import astropy.time as t
+import astropy.units as u
 import numpy as np
 import voluptuous as vol
-from astropy import time as t
-from astropy import units as u
 from astropy.io import fits
 from astropy.io.fits.hdu.hdulist import HDUList
 from dkist_fits_specifications import spec214
 from dkist_fits_specifications.utils import expand_naxis
-from numpy import ndarray
+from voluptuous.error import Invalid
 
 from dkist_header_validator.exceptions import ReturnTypeException
 from dkist_header_validator.exceptions import SpecSchemaDefinitionException
@@ -33,6 +37,65 @@ logger = logging.getLogger(__name__)
 __all__ = ["SpecValidator", "SpecSchema"]
 
 
+class FormatInvalid(Invalid):
+    """
+    An value which does not match the schema format.
+    """
+
+
+class FormatValidator:
+    """
+    Validate that the format is valid.
+    """
+
+    supported_formats = ["isot", "unit"]
+
+    def __init__(self, format_value):
+        if format_value not in self.supported_formats:
+            raise SpecSchemaDefinitionException(f"{format_value} is an unknown format to validate.")
+        self.format_value = format_value
+
+    def __call__(self, value):
+        if self.format_value == "unit":
+            try:
+                return u.Unit(value, format="fits")
+            except Exception:
+                raise FormatInvalid(f"{value} is not a valid FITS unit")
+
+        if self.format_value == "isot":
+            try:
+                return t.Time(value, format="fits")
+            except Exception:
+                raise FormatInvalid(f"{value} is not a valid FITS time")
+
+    def __repr__(self):
+        return f"FormatValidator(format='{self.format_value}')"
+
+
+class FITSFloatInvalid(Invalid):
+    """
+    Validates a header value is a valid FITS float.
+
+    This means it's not NaN or ±Inf.
+    """
+
+
+class FITSFloatValidsator:
+    """
+    Validate a FITS floating point value.
+
+    This validator assumes the type has also been validated independently.
+    """
+
+    def __call__(self, value):
+        if not np.isfinite(value):
+            raise FITSFloatInvalid(
+                f"{value} is not finite, floats in a FITS header must be finite."
+            )
+
+        return value
+
+
 class SpecSchema:
     """
     Define a schema that is used to validate FITS headers for a spec based upon structured definitions
@@ -132,25 +195,22 @@ class SpecSchema:
         :param key_schema: Spec schema (from yml) used to generate voluptuous schema
         :return: Voluptuous schema for a key
         """
-        type_map = {"int": int, "float": float, "str": str, "bool": bool}
-        if key_schema.get("values"):
-            return vol.All(type_map[key_schema.get("type")], vol.Any(*key_schema.get("values")))
-        return vol.All(type_map[key_schema.get("type")])
+        type_map = {"int": Integral, "float": Real, "str": str, "bool": bool}
 
-    @staticmethod
-    def _validate_units(schema: dict, headers: dict):
-        """
-        Validates units of headers against the spec raising
-        ValueErrors upon failure.
-        :param schema: Definition of the spec's schema
-        :param headers: Fits file headers
-        """
-        for key, value in schema.items():
-            if key in headers:
-                if value.get("format", None) == "unit":
-                    return u.Unit(headers[key], format="fits")
-                if value.get("format", None) == "isot":
-                    return t.Time(headers[key], format="fits")
+        checks = []
+        # Always check type
+        checks.append(type_map[key_schema.get("type")])
+
+        if key_schema.get("values"):
+            checks.append(vol.Any(*key_schema.get("values")))
+
+        if "format" in key_schema:
+            checks.append(FormatValidator(key_schema["format"]))
+
+        if key_schema.get("type") == "float":
+            checks.append(FITSFloatValidsator())
+
+        return vol.All(*checks)
 
     def _add_keys_to_schema(self, schema: dict) -> dict:
         schema_keys = {}  # keys to be added to voluptuous schema
@@ -186,7 +246,6 @@ class SpecSchema:
         spec_schema = {}
 
         schema = self._check_for_expansion(headers)
-        self._validate_units(schema, headers)
         spec_schema.update(self._add_keys_to_schema(schema))
         if extra:
             return vol.Schema(spec_schema, extra=vol.ALLOW_EXTRA)
@@ -262,14 +321,40 @@ class SpecValidator:
         try:
             self.spec_schema(headers, extra)
         except vol.MultipleInvalid as e:
-            validation_errors = {
-                error.path[0]: f"{error.msg}. Actual value: "
-                f"{headers.get(error.path[0], 'Required keyword not present')}"
-                for error in e.errors
-            }
+            for error in e.errors:
+                value_str = "Required keyword not present"
+                if error.path[0] in headers:
+                    value_str = (
+                        f"Actual value: {headers[error.path[0]]!r} {type(headers[error.path[0]])}"
+                    )
+                validation_errors[error.path[0]] = f"{error.msg}. {value_str}"
         # Raise exception if we have errors
         if validation_errors:
-            logger.error(f"Errors during validation: errors={validation_errors}")
+            missing_list = []
+            badtype_list = []
+            other_list = []
+
+            for key in validation_errors.keys():
+                if "required key not provided" in validation_errors[key]:
+                    missing_list.append(key)
+                elif "expected" in validation_errors[key]:
+                    badtype_list.append(key)
+                else:
+                    other_list.append(key)
+            # Log the bad keys
+            for sublist, message in zip(
+                [missing_list, badtype_list, other_list],
+                [
+                    "The following keys are missing:",
+                    "The following keys have bad types:",
+                    "The following keys have other errors:",
+                ],
+            ):
+                if len(sublist) > 0:
+                    logger.error(f"\n{message}")
+                    for k in sorted(sublist):
+                        logger.error(f"{str(k):<10}:\t{validation_errors[k]}")
+
             raise self.SchemaValidationException(errors=validation_errors)
         logger.debug("Schema validation succeeded")
 
@@ -277,7 +362,7 @@ class SpecValidator:
 
     def _validate_headers(
         self, input_headers: Union[HDUList, dict, fits.header.Header], extra
-    ) -> dict:
+    ) -> Tuple[dict, dict]:
         """
         Validates open input headers against the instance spec_schema
         :param input_headers: The input headers to validate in the following formats:
@@ -287,11 +372,17 @@ class SpecValidator:
         :param extra: switch for validation to allow extra keys in schema
         :return: dictionary of verified headers to be used later
         """
+        if isinstance(input_headers, HDUList):
+            if len(input_headers) > 2:
+                raise ValidationException(
+                    "Too many HDUs in your HDUList! May only have two HDUs at most."
+                )
         headers = self._headers_to_dict(input_headers)
+        fits_cards = self._capture_fits_cards(headers)
         verified_headers = self.verify_headers(headers, extra)
-        return verified_headers
+        return verified_headers, fits_cards
 
-    def _validate_file(self, input_headers: Union[str, IO], extra) -> dict:
+    def _validate_file(self, input_headers: Union[str, IO], extra) -> Tuple[dict, dict]:
         """
         Validates files against the astropy and then instance spec_schema
         :param input_headers: The input headers to validate in the following formats:
@@ -301,54 +392,78 @@ class SpecValidator:
         :return: dictionary of verified headers to be used later
         """
         try:
-            with fits.open(input_headers, disable_image_compression=True) as hdul:
+            with fits.open(input_headers) as hdul:
+                if len(hdul) > 2:
+                    raise ValidationException(
+                        "Too many HDUs in your HDUList! May only have two HDUs at most."
+                    )
                 # verify fits headers with astropy verify library
                 hdul.verify("exception")
                 # normalize headers into a dict
-                hdus = self._headers_to_dict(hdul[0].header)
+                try:
+                    hdus = self._headers_to_dict(hdul[1].header)
+                except IndexError:  # non-compressed
+                    hdus = self._headers_to_dict(hdul[0].header)
+                fits_cards = self._capture_fits_cards(hdus)
                 verified_headers = self.verify_headers(hdus, extra)
-                return verified_headers
+                return verified_headers, fits_cards
         except (ValueError, FileNotFoundError, OSError, IndexError) as exc:
             logger.error(f"Cannot parse headers: detail = {exc}")
             raise ValidationException(f"Cannot parse headers", errors={type(exc): str(exc)})
 
     @staticmethod
-    def _return_hdulist(validated_headers) -> HDUList:
+    def _return_hdulist(validated_headers, fits_cards) -> HDUList:
         """
         Returns validated headers as an HDUList
-        :param validated_headers: Already validated/translated headers to be written out into an HDUList
+        :param validated_headers: Already validated/translated headers to be written out into
+          an HDUList
+        :param fits_cards: Any special cards to be included in the HDUList
         :return: HDUList
         """
         if isinstance(validated_headers, HDUList):
             return validated_headers
-        temp_array: ndarray = np.ones((1, 1, 1), dtype=np.int16)
+        temp_array: np.ndarray = np.ones((1, 1, 1), dtype=np.int16)
         new_hdu = fits.PrimaryHDU(temp_array)
         for (key, value) in validated_headers.items():
             new_hdu.header[key] = value
+        for key in fits_cards:
+            new_hdu.header[key] = str(fits_cards[key])
         new_hdu_list = fits.HDUList([new_hdu])
         return new_hdu_list
 
     @staticmethod
-    def _return_dictionary(validated_headers) -> dict:
+    def _return_dictionary(validated_headers, fits_cards) -> dict:
         """
         Returns validated headers as a dictionary
-        :param validated_headers: Already validated/translated headers to be written out into a dictionary
+        :param validated_headers: Already validated/translated headers to be written out into
+          a dictionary
+        :param fits_cards: Any special cards to be included in the dictionary
         :return: dictionary
         """
+        for key in fits_cards:
+            validated_headers[key] = str(fits_cards[key])
         return validated_headers
 
     @staticmethod
-    def _return_BytesIO(validated_headers, input_headers) -> BytesIO:
+    def _return_BytesIO(validated_headers, input_headers, fits_cards) -> BytesIO:
         """
         Returns validated headers as a BytesIO object
-        :param validated_headers: Already validated/translated headers to be written out into a dictionary
+        :param validated_headers: Already validated/translated headers to be written out into
+          the BytesIO object
+        :param input_headers: original filepath
+        :param fits_cards: Any special cards to be included in the BytesIO object
         :return: BytesIO object
         """
-        if input_headers:
-            hdul = fits.open(input_headers, disable_image_compression=True)
-            new_hdu = fits.PrimaryHDU(hdul[0].data)
+        if isinstance(input_headers, pathlib.PosixPath):
+            hdul = fits.open(input_headers)
+            if hdul[0].data == None:  # check if data is in second HDU
+                new_hdu = fits.PrimaryHDU(hdul[1].data)
+            else:
+                new_hdu = fits.PrimaryHDU(hdul[0].data)
             for (key, value) in validated_headers.items():
                 new_hdu.header[key] = value
+            for key in fits_cards:
+                new_hdu.header[key] = str(fits_cards[key])
             new_hdu_list = fits.HDUList([new_hdu])
             return BytesIO(
                 new_hdu_list.writeto(
@@ -361,47 +476,62 @@ class SpecValidator:
         raise ReturnTypeException("No data. Cannot write BytesIO object.")
 
     @staticmethod
-    def _return_file(validated_headers, input_headers) -> (str, IO):
+    def _return_file(validated_headers, input_headers, fits_cards) -> (str, IO):
         """
         Returns validated headers as a FITS file
-        :param validated_headers: Already validated/translated headers to be written out into a FITS file
+        :param validated_headers: Already validated/translated headers to be written out
+          into a FITS file
         :param input_headers: original filepath
+        :param fits_cards: Any special cards to be included in the FITS file
         :return: FITS file
         """
-        """hdul = fits.open(input_headers, disable_image_compression=True)
-        new_hdu = fits.PrimaryHDU(hdul[0].data)
-        for (key, value) in validated_headers.items():
-            new_hdu.header[key] = value
-        new_hdu_list = fits.HDUList([new_hdu])
-        new_hdu_list.writeto(
-            str(os.path.basename(input_headers)),
-            overwrite=True,
-            output_verify="exception",
-            checksum=True,
-        )"""
-        if input_headers:
-            hdul = fits.open(input_headers, disable_image_compression=True)
-            new_hdu = fits.PrimaryHDU(hdul[0].data)
+        if isinstance(input_headers, pathlib.PosixPath):
+            hdul = fits.open(input_headers)
+            if hdul[0].data == None:  # check if data is in second HDU
+                new_hdu = fits.PrimaryHDU(hdul[1].data)
+            else:
+                new_hdu = fits.PrimaryHDU(hdul[0].data)
             for (key, value) in validated_headers.items():
                 new_hdu.header[key] = value
+            for key in fits_cards:
+                new_hdu.header[key] = str(fits_cards[key])
             new_hdu_list = fits.HDUList([new_hdu])
-            return new_hdu_list.writeto(
+            new_hdu_list.writeto(
                 str(os.path.basename(input_headers)),
                 overwrite=True,
                 output_verify="exception",
                 checksum=True,
             )
+            return os.path.basename(input_headers)
         raise ReturnTypeException("No data. Cannot write file.")
 
-    def _format_output(self, return_type, validated_headers, input_headers=None):
+    def _format_output(self, return_type, validated_headers, input_headers=None, fits_cards=None):
+        fits_cards = fits_cards or {}
         if return_type == Path:
-            return self._return_file(validated_headers, input_headers)
+            return self._return_file(validated_headers, input_headers, fits_cards)
         if return_type == BytesIO:
-            return self._return_BytesIO(validated_headers, input_headers)
+            return self._return_BytesIO(validated_headers, input_headers, fits_cards)
         if return_type == dict:
-            return self._return_dictionary(validated_headers)
+            return self._return_dictionary(validated_headers, fits_cards)
         if return_type == HDUList or fits.header.Header:
-            return self._return_hdulist(validated_headers)
+            return self._return_hdulist(validated_headers, fits_cards)
+
+    def _capture_fits_cards(self, validated_headers) -> dict:
+        """
+        Pull special fits cards out of validated_headers dict.
+        This is necessary for astropy header formatting.
+
+        :param validated_headers: validated headers
+        :return: fits_cards: dictionary containing special fits header keys and values
+        """
+        fits_cards = {}
+        if "HISTORY" in validated_headers:
+            fits_cards["HISTORY"] = str(validated_headers["HISTORY"])
+            validated_headers.pop("HISTORY")
+        if "COMMENT" in validated_headers:
+            fits_cards["COMMENT"] = str(validated_headers["COMMENT"])
+            validated_headers.pop("COMMENT")
+        return fits_cards
 
     def validate(self, input_headers, return_type=HDUList, extra=True):
         """
@@ -424,11 +554,10 @@ class SpecValidator:
         :raises: SpecValidationException or subclass
         """
         if isinstance(input_headers, (dict, fits.header.Header, HDUList)):
-            validated_headers = self._validate_headers(input_headers, extra)
-            return self._format_output(return_type, validated_headers)
-        else:
-            validated_headers = self._validate_file(input_headers, extra)
-            return self._format_output(return_type, validated_headers, input_headers)
+            validated_headers, fits_cards = self._validate_headers(input_headers, extra)
+            return self._format_output(return_type, validated_headers, input_headers, fits_cards)
+        validated_headers, fits_cards = self._validate_file(input_headers, extra)
+        return self._format_output(return_type, validated_headers, input_headers, fits_cards)
 
     def validate_and_translate(self, input_headers, return_type=HDUList, extra=True):
         """
@@ -451,10 +580,10 @@ class SpecValidator:
         :raises: SpecValidationException or subclass
         """
         if isinstance(input_headers, (dict, fits.header.Header, HDUList)):
-            validated_headers = self._validate_headers(input_headers, extra)
+            validated_headers, fits_cards = self._validate_headers(input_headers, extra)
             translated_headers = convert_spec122_to_spec214(validated_headers)
-            return self._format_output(return_type, translated_headers)
+            return self._format_output(return_type, translated_headers, fits_cards)
         else:
-            validated_headers = self._validate_file(input_headers, extra)
+            validated_headers, fits_cards = self._validate_file(input_headers, extra)
             translated_headers = convert_spec122_to_spec214(validated_headers)
-            return self._format_output(return_type, translated_headers, input_headers)
+            return self._format_output(return_type, translated_headers, input_headers, fits_cards)
