@@ -436,7 +436,7 @@ def valid_test_header_params(tmpdir_factory):
     """
     temp_dir = tmpdir_factory.mktemp("valid test_headers_temp")
     file_name = temp_dir.join("tmp_fits_file.fits")
-    valid_array = np.ones((1, 1, 1), dtype=np.float)
+    valid_array = np.ones((1, 1, 1), dtype=float)
     valid_hdu = fits.PrimaryHDU(valid_array)
     valid_hdu_list = fits.HDUList([valid_hdu])
     valid_hdu_list.writeto(str(file_name))
@@ -473,7 +473,7 @@ def valid_test_compressed_file(tmpdir_factory):
     """
     temp_dir = tmpdir_factory.mktemp("valid test_headers_temp")
     file_name = temp_dir.join("tmp_fits_file.fits")
-    valid_array = np.ones((1, 1, 1), dtype=np.float)
+    valid_array = np.ones((1, 1, 1), dtype=float)
     primary_hdu = fits.PrimaryHDU()
     valid_hdu = fits.CompImageHDU(valid_array)
     valid_hdu_list = fits.HDUList([primary_hdu, valid_hdu])
@@ -510,7 +510,7 @@ def invalid_test_header_params(tmpdir_factory):
     """
     temp_dir = tmpdir_factory.mktemp("invalid_test_headers_temp")
     file_name = temp_dir.join("tmp_fits_file.fits")
-    invalid_array = np.ones((1, 1), dtype=np.float)
+    invalid_array = np.ones((1, 1), dtype=float)
     primary_hdu = fits.PrimaryHDU()
     invalid_hdu = fits.CompImageHDU(invalid_array)
     invalid_hdu_list = fits.HDUList([primary_hdu, invalid_hdu])
