@@ -1,4 +1,5 @@
 import logging
+from datetime import datetime
 from typing import Any
 from typing import Dict
 from typing import IO
@@ -9,6 +10,9 @@ from typing import Union
 
 from astropy.io import fits
 from astropy.io.fits.hdu.hdulist import HDUList
+from dkist_fits_specifications.spec214 import expand_index_d
+from dkist_fits_specifications.spec214 import expand_index_k
+from dkist_fits_specifications.spec214 import level0
 from dkist_fits_specifications.spec214 import load_full_spec214
 from dkist_fits_specifications.utils import expand_naxis
 
@@ -31,6 +35,15 @@ def convert_spec122_to_spec214(
     input_headers, input_data = _parse_spec122(spec122_input)
     # convert headers
     output_headers = _add_214_headers(input_headers)
+    # update DATE keyword
+    output_headers["DATE"] = datetime.now().isoformat()
+    # remove BZERO and BSCALE keywords- we don't use these anymore
+    try:
+        del output_headers["BZERO"]
+        del output_headers["BSCALE"]
+    except KeyError:
+        # if they aren't there we can't delete them
+        pass
     if input_data:  # return hdu list if the input had data
         return _format_output_hdu(output_headers, input_data)
     return output_headers  # return headers if only headers were given
@@ -54,10 +67,10 @@ def _parse_spec122(
 
     # If headers are of any other type, see if it is a file and try to open that
     try:  # compressed
-        with fits.open(spec122_input, disable_image_compression=True) as hdus:
+        with fits.open(spec122_input) as hdus:
             return dict(hdus[1].header), hdus[1].data
     except IndexError:  # non-compressed
-        with fits.open(spec122_input, disable_image_compression=True) as hdus:
+        with fits.open(spec122_input) as hdus:
             return dict(hdus[0].header), hdus[0].data
 
 
@@ -69,14 +82,15 @@ def _format_output_hdu(hdr, data) -> HDUList:
     return hdu_list
 
 
-generic_spec214_schema = load_full_spec214().values()  # global for performance
+generic_spec214_schema = level0.load_level0_spec214().values()
 
 
 def _create_spec214_schema_for_header(hdr: dict) -> Dict[str, Dict[str, Any]]:
     s = {
         key: schema for definition in generic_spec214_schema for (key, schema) in definition.items()
     }
-    return expand_naxis(hdr["NAXIS"], s)  # only need to account for spec 122 input
+    hdr["DNAXIS"] = hdr["NAXIS"]
+    return expand_index_d(expand_naxis(hdr["NAXIS"], s), DNAXIS=hdr["DNAXIS"])
 
 
 def _add_214_headers(hdr: dict):
